from .kili import clean_json_response, convert_to_pixel_coords, format_json_response
from .format.llm import (
    convert_from_kili_to_llm_rlhf_format,
    convert_from_kili_to_llm_static_or_dynamic_format,
)
from .format.voc import convert_from_kili_to_voc_format
from .format.yolo import convert_from_kili_to_yolo_format
from .format.coco import convert_from_kili_to_coco_format

__all__ = [
    clean_json_response,
    convert_from_kili_to_coco_format,
    convert_from_kili_to_llm_rlhf_format,
    convert_from_kili_to_llm_static_or_dynamic_format,
    convert_from_kili_to_voc_format,
    convert_from_kili_to_yolo_format,
    convert_to_pixel_coords,
    format_json_response,
]



from typing import Dict, List


def scale_all_vertices(object_, width: int, height: int):
    if isinstance(object_, List):
        return [scale_all_vertices(obj, width=width, height=height) for obj in object_]

    if isinstance(object_, Dict):
        if sorted(object_.keys()) == ["x", "y"]:
            return _scale_vertex(object_, width=width, height=height)
        return {
            key: scale_all_vertices(value, width=width, height=height)
            for key, value in object_.items()
        }
    return object_


def get_asset_dimensions(asset: Dict):
    """Returns the width and height of the asset."""
    resolution = asset["resolution"]
    return resolution["width"], resolution["height"]


def _scale_vertex(vertex: Dict, width: int, height: int) -> Dict:
    return {"x": vertex["x"] * width, "y": vertex["y"] * height}
 


import pytest

from kili_formats.exceptions import NotCompatibleOptions
from kili_formats.kili import convert_to_pixel_coords
from .fakes.image import image_asset, image_asset_rotated, image_project, image_project_asset_normalized, image_project_asset_unnormalized, image_rotated_project_asset_unnormalized
from .fakes.pdf import pdf_asset, pdf_asset_rotated, pdf_project, pdf_project_asset_unnormalized
from .fakes.video import video_asset, video_project, video_project_asset_unnormalized

def test_kili_convert_to_pixel_coords_pdf():
    """Test the conversion of coordinates from normalized to pixel values for PDF files."""
    scaled_asset = convert_to_pixel_coords(pdf_asset, pdf_project, normalized_coordinates=False)
    assert scaled_asset == pdf_project_asset_unnormalized

def test_kili_convert_to_pixel_coords_pdf_rotated_throw_error():
    """Test the conversion of coordinates from normalized to pixel values for PDF files raises error if rotated."""
    with pytest.raises(NotCompatibleOptions, match="PDF labels cannot be rotated"):
        convert_to_pixel_coords(pdf_asset_rotated, pdf_project, normalized_coordinates=False)

def test_kili_convert_to_pixel_coords_image():
    """Test the conversion of coordinates from normalized to pixel values for image files."""
    scaled_asset = convert_to_pixel_coords(image_asset, image_project, normalized_coordinates=False)
    assert scaled_asset == image_project_asset_unnormalized

def test_kili_convert_to_pixel_coords_image_with_normalized_coordinates():
    """Test the conversion of coordinates from normalized to pixel values for image files."""
    scaled_asset = convert_to_pixel_coords(image_asset, image_project, normalized_coordinates=True)
    assert scaled_asset == image_project_asset_normalized

def test_kili_convert_to_pixel_coords_video():
    """Test the conversion of coordinates from normalized to pixel values for video files."""
    scaled_asset = convert_to_pixel_coords(video_asset, video_project, normalized_coordinates=False)
    assert scaled_asset == video_project_asset_unnormalized

def test_kili_convert_to_pixel_coords_image_rotated():
    """Test the conversion of coordinates from normalized to pixel values for rotated image files."""
    scaled_asset = convert_to_pixel_coords(
        image_asset_rotated, image_project, normalized_coordinates=False
    )
    assert scaled_asset == image_rotated_project_asset_unnormalized



from pathlib import Path
import json
from typing import Dict, List, Optional, Union, cast

from ..types import ChatItemRole, ConversationLabel, ExportLLMItem, JobLevel


DEFAULT_JOB_LEVEL = JobLevel.ROUND
SEPARATOR = "___"

def convert_from_kili_to_llm_static_or_dynamic_format(annotations: List[Dict], chat_items: List[Dict], jobs: Dict) -> ConversationLabel:
    formatted_label = {JobLevel.COMPLETION: {}, JobLevel.CONVERSATION: {}, JobLevel.ROUND: {}}

    for job_name, job in jobs.items():
        job_level = job.get("level", DEFAULT_JOB_LEVEL)

        if job_level == JobLevel.COMPLETION:
            job_label = format_completion_job(job_name, annotations, chat_items)
        elif job_level == JobLevel.CONVERSATION:
            job_label = format_conversation_job(job_name, annotations)
        elif job_level == JobLevel.ROUND:
            job_label = format_round_job(job_name, annotations, chat_items)
        else:
            raise ValueError(f"Unknown job level: {job_level}")

        if job_label:
            formatted_label[job_level][job_name] = job_label

    return cast(ConversationLabel, formatted_label)

def format_completion_job(
    job_name: str, annotations: List[Dict], chat_items: List[Dict]
) -> Dict:
    id_to_external_id = {
        item["id"]: item.get("externalId") or item["id"] for item in chat_items
    }
    job_annotations = [
        annotation for annotation in annotations if annotation.get("job") == job_name
    ]
    return {
        id_to_external_id.get(annotation["chatItemId"]): annotation["annotationValue"]
        for annotation in job_annotations
    }

def format_round_job(
    job_name: str, annotations: List[Dict], chat_items: List[Dict]
) -> Dict:
    id_to_external_id = {
        item["id"]: item.get("externalId") or item["id"] for item in chat_items
    }
    job_annotations = [
        annotation for annotation in annotations if annotation.get("job") == job_name
    ]
    user_chat_item_ids = [
        chat_item.get("external_id") or chat_item["id"]
        for chat_item in chat_items
        if chat_item["role"] == ChatItemRole.USER
    ]

    return {
        user_chat_item_ids.index(annotation["chatItemId"]): format_annotation_value(
            annotation["annotationValue"], id_to_external_id
        )
        for annotation in job_annotations
    }

def format_conversation_job(job_name: str, annotations: List[Dict]) -> Dict:
    annotation = next(
        (annotation for annotation in annotations if annotation["job"] == job_name), None
    )
    if annotation:
        return annotation["annotationValue"]
    return {}

def format_annotation_value(annotation_value: Dict, id_to_external_id: Dict[str, str]) -> Dict:
    if annotation_value.get("choice"):
        return {
            "code": annotation_value["choice"]["code"],
            "firstId": id_to_external_id.get(annotation_value["choice"]["firstId"]),
            "secondId": id_to_external_id.get(annotation_value["choice"]["secondId"]),
        }
    return annotation_value

def convert_from_kili_to_llm_rlhf_format(
    assets: List[Dict], json_interface: Dict[str, Dict], logging
) -> List[Dict[str, Union[List[str], str]]]:
    result = []
    for asset in assets:
        result.append(
            {
                "raw_data": _format_raw_data(asset),
                "status": asset["status"],
                "external_id": asset["externalId"],
                "metadata": asset["jsonMetadata"],
                "labels": list(
                    map(
                        lambda label: {
                            "author": label["author"]["email"],
                            "created_at": label["createdAt"],
                            "label_type": label["labelType"],
                            "label": _format_json_response(
                                json_interface["jobs"], label["jsonResponse"], logging
                            ),
                        },
                        asset["labels"],
                    )
                ),
            }
        )
    return result

def _format_json_response(
    jobs_config: Dict, json_response: Dict, logging
) -> Dict[str, Dict[str, Union[str, List[str]]]]:
    result = {}
    for job_name, job_value in json_response.items():
        job_config = jobs_config[job_name]
        if job_config is None:
            continue
        if job_config["mlTask"] == "CLASSIFICATION":
            result[job_name] = []
            for category in job_value["categories"]:
                result[job_name].append(category["name"])
                if "children" in category:
                    for child_name, child_value in _format_json_response(
                        jobs_config, category["children"], logging
                    ).items():
                        result[f"{job_name}.{category['name']}.{child_name}"] = child_value
        elif job_config["mlTask"] == "TRANSCRIPTION":
            result[job_name] = job_value["text"]
        else:
            logging.warning(f"Job {job_name} with mlTask {job_config['mlTask']} not supported")
    if len(result) != 0:
        return {"conversation": result}
    return result


def _format_raw_data(
    asset, step_number: Optional[int] = None, all_model_keys: Optional[bool] = False
) -> List[ExportLLMItem]:
    raw_data = []

    chat_id = asset["jsonMetadata"].get("chat_id", None)

    if (
        "chat_item_ids" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["chat_item_ids"], str)
        and len(asset["jsonMetadata"]["chat_item_ids"]) > 0
    ):
        chat_items_ids = str.split(asset["jsonMetadata"]["chat_item_ids"], SEPARATOR)
        if step_number is not None:
            chat_items_ids = chat_items_ids[step_number * 3 :]
    else:
        chat_items_ids = []

    if (
        "models" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["models"], str)
        and len(asset["jsonMetadata"]["models"]) > 0
    ):
        models = str.split(asset["jsonMetadata"]["models"], SEPARATOR)
    else:
        models = []
    with Path(asset["content"]).open(encoding="utf8") as file:
        data = json.load(file)
    version = data.get("version", None)
    if version == "0.1":
        prompts = data["prompts"]
        if step_number is not None:
            prompts = [prompts[step_number]]
        for index, prompt in enumerate(prompts):
            raw_data.append(
                ExportLLMItem(
                    {
                        "role": prompt.get("title", "user"),
                        "content": prompt["prompt"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": None,
                    }
                )
            )
            raw_data.extend(
                ExportLLMItem(
                    {
                        "role": completion.get("title", "assistant"),
                        "content": completion["content"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": models[index_completion]
                        if (
                            (index == len(prompts) - 1 or all_model_keys)
                            and len(models) > index_completion
                        )
                        else None,
                    }
                )
                for index_completion, completion in enumerate(prompt["completions"])
            )
    else:
        raise ValueError(f"Version {version} not supported")
    return raw_data

def _safe_pop(lst, index=0):
    try:
        return lst.pop(index)
    except IndexError:
        return None


import os
import tempfile

from kili_formats.format.llm import convert_from_kili_to_llm_rlhf_format, convert_from_kili_to_llm_static_or_dynamic_format
from .fakes.llm_dynamic import llm_dynamic_annotations, llm_dynamic_chat_items, llm_dynamic_expected_label, jobs
from .fakes.llm_static import llm_static_annotations, llm_static_chat_items, llm_static_expected_result
from .fakes.llm_rlhf import llm_rlhf_expected_result, llm_rlhf_assets, llm_rlhf_json_interface, mock_raw_asset_content


def test_convert_from_kili_to_llm_dynamic_format():
    """Test the conversion from Kili format to LLM dynamic format."""

    formatted_label = convert_from_kili_to_llm_static_or_dynamic_format(
        llm_dynamic_annotations, llm_dynamic_chat_items, jobs
    )

    assert formatted_label == llm_dynamic_expected_label

def test_convert_from_kili_to_llm_static_format():
    """Test the conversion from Kili format to LLM static format."""

    formatted_label = convert_from_kili_to_llm_static_or_dynamic_format(
        llm_static_annotations, llm_static_chat_items, jobs
    )

    assert formatted_label == llm_static_expected_result

def test_convert_from_kili_to_llm_rlhf_format():
    """Test the conversion from Kili format to LLM static format."""
    
    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp:
            tmp.write(mock_raw_asset_content)
        for mocked_asset in llm_rlhf_assets:
            mocked_asset["content"] = path
        print("mocked_asset", mocked_asset["content"])
        formatted_label = convert_from_kili_to_llm_rlhf_format(llm_rlhf_assets, llm_rlhf_json_interface, logging=None)
        assert formatted_label == llm_rlhf_expected_result
    finally:
        os.remove(path)


from typing import Dict, Sequence
import xml.etree.ElementTree as ET
from xml.dom import minidom

from kili_formats.tool.base import reverse_rotation_vertices


def convert_from_kili_to_voc_format(response: Dict, width: int, height: int, parameters: Dict, valid_jobs: Sequence[str]) -> str:
    xml_label = ET.Element("annotation")

    _provide_voc_headers(xml_label, width, height, parameters=parameters)

    _parse_annotations(response, xml_label, width, height, valid_jobs)

    xmlstr = minidom.parseString(ET.tostring(xml_label)).toprettyxml(indent="   ")

    return xmlstr

def _provide_voc_headers(xml_label: ET.Element, width: int, height: int, parameters: Dict) -> None:
    folder = ET.SubElement(xml_label, "folder")
    folder.text = parameters.get("folder", "")

    filename = ET.SubElement(xml_label, "filename")
    filename.text = parameters.get("filename", "")

    path = ET.SubElement(xml_label, "path")
    path.text = parameters.get("path", "")

    source = ET.SubElement(xml_label, "source")
    database = ET.SubElement(source, "database")
    database.text = "Kili Technology"

    size = ET.SubElement(xml_label, "size")
    width_xml = ET.SubElement(size, "width")
    width_xml.text = str(width)
    height_xml = ET.SubElement(size, "height")
    height_xml.text = str(height)
    depth = ET.SubElement(size, "depth")
    depth.text = parameters.get("depth", "3")

    segmented = ET.SubElement(xml_label, "segmented")
    segmented.text = 0  # type: ignore

def _parse_annotations(response: Dict, xml_label: ET.Element, width: int, height:int, valid_jobs: Sequence[str]) -> None:
    # pylint: disable=too-many-locals
    rotation_val = 0
    if "ROTATION_JOB" in response:
        rotation_val = response["ROTATION_JOB"]["rotation"]
    for job_name, job_response in response.items():
        if job_name not in valid_jobs:
            continue
        if "annotations" in job_response:
            annotations = job_response["annotations"]
            for annotation in annotations:
                vertices = annotation["boundingPoly"][0]["normalizedVertices"]
                categories = annotation["categories"]
                for category in categories:
                    annotation_category = ET.SubElement(xml_label, "object")
                    name = ET.SubElement(annotation_category, "name")
                    name.text = category["name"]
                    job_name_xml = ET.SubElement(annotation_category, "job_name")
                    job_name_xml.text = job_name
                    pose = ET.SubElement(annotation_category, "pose")
                    pose.text = "Unspecified"
                    truncated = ET.SubElement(annotation_category, "truncated")
                    truncated.text = "0"
                    difficult = ET.SubElement(annotation_category, "difficult")
                    difficult.text = "0"
                    occluded = ET.SubElement(annotation_category, "occluded")
                    occluded.text = "0"
                    bndbox = ET.SubElement(annotation_category, "bndbox")
                    vertices_before_rotate = reverse_rotation_vertices(vertices, rotation_val)
                    x_vertices = [v["x"] * width for v in vertices_before_rotate]
                    y_vertices = [v["y"] * height for v in vertices_before_rotate]
                    xmin = ET.SubElement(bndbox, "xmin")
                    xmin.text = str(round(min(x_vertices)))
                    xmax = ET.SubElement(bndbox, "xmax")
                    xmax.text = str(round(max(x_vertices)))
                    ymin = ET.SubElement(bndbox, "ymin")
                    ymin.text = str(round(min(y_vertices)))
                    ymax = ET.SubElement(bndbox, "ymax")
                    ymax.text = str(round(max(y_vertices)))


from kili_formats.format.voc import convert_from_kili_to_voc_format

xml_str = '<?xml version="1.0" ?>\n<annotation>\n   <folder/>\n   <filename>filename.xml</filename>\n   <path/>\n   <source>\n      <database>Kili Technology</database>\n   </source>\n   <size>\n      <width>1920</width>\n      <height>1080</height>\n      <depth>3</depth>\n   </size>\n   <segmented/>\n   <object>\n      <name>OBJECT_A</name>\n      <job_name>JOB_0</job_name>\n      <pose>Unspecified</pose>\n      <truncated>0</truncated>\n      <difficult>0</difficult>\n      <occluded>0</occluded>\n      <bndbox>\n         <xmin>317</xmin>\n         <xmax>1609</xmax>\n         <ymin>281</ymin>\n         <ymax>863</ymax>\n      </bndbox>\n   </object>\n</annotation>\n'
def test_voc_convert_from_kili_to_voc_format():
    """Test the conversion from Kili format to VOC format."""

    response = {
        "JOB_0": {
        "annotations": [
            {
                "categories": [{"confidence": 100, "name": "OBJECT_A"}],
                "jobName": "JOB_0",
                "mid": "2022040515434712-7532",
                "mlTask": "OBJECT_DETECTION",
                "boundingPoly": [
                    {
                        "normalizedVertices": [
                            {"x": 0.16504140348233334, "y": 0.7986938935103378},
                            {"x": 0.16504140348233334, "y": 0.2605618833516984},
                            {"x": 0.8377886490672706, "y": 0.2605618833516984},
                            {"x": 0.8377886490672706, "y": 0.7986938935103378},
                        ]
                    }
                ],
                "type": "rectangle",
                "children": {},
            }
        ]
    }
    }

    width = 1920
    height = 1080
    parameters = {
        "filename": "filename.xml",
    }

    xmlstr = convert_from_kili_to_voc_format(response, width, height, parameters, ["JOB_0"])
    assert xmlstr == xml_str


import json
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from kili_formats.media.image import get_frame_dimensions, get_image_dimensions
from kili_formats.media.video import cut_video, get_video_dimensions
from kili_formats.tool.base import reverse_rotation_vertices
from kili_formats.types import Job
import numpy as np
from shapely.geometry import Polygon
from shapely.ops import polygonize
from shapely.validation import make_valid


from .types import (
    CocoAnnotation,
    CocoAnnotationModifier,
    CocoCategory,
    CocoFormat,
    CocoImage,
)

DATA_SUBDIR = "data"


# pylint: disable=too-many-arguments
def convert_from_kili_to_coco_format(
    jobs: Dict[str, Job],
    assets: List[Dict],
    output_dir: Path,
    title: str,
    project_input_type: str,
    annotation_modifier: Optional[CocoAnnotationModifier],
    merged: bool,
) -> Tuple[CocoFormat, List[Path]]:
    """Creates the following structure on the disk.

    <dataset_dir>/
        data/
            <filename0>.<ext>
            <filename1>.<ext>
            ...
        labels.json.

    We iterate on the assets and create a coco format for each asset.

    Note: the jobs should only contains elligible jobs.
    """
    infos_coco = {
        "year": time.strftime("%Y"),
        "version": "1.0",
        "description": f"{title} - Exported from Kili Python Client",
        "contributor": "Kili Technology",
        "url": "https://kili-technology.com",
        "date_created": datetime.now().isoformat(),
    }
    labels_json = CocoFormat(
        info=infos_coco,
        licenses=[],
        categories=[],
        images=[],
        annotations=[],
    )

    # Mapping category - category id
    cat_kili_id_to_coco_id, labels_json["categories"] = _get_coco_categories_with_mapping(
        jobs, merged
    )

    if merged:
        labels_json["images"], labels_json["annotations"] = _get_coco_images_and_annotations(
            jobs,
            assets,
            cat_kili_id_to_coco_id,
            project_input_type,
            annotation_modifier,
            is_single_job=False,
        )

        label_file_name = output_dir / "labels.json"
        label_file_name.parent.mkdir(parents=True, exist_ok=True)
        with label_file_name.open("w") as outfile:
            json.dump(labels_json, outfile)
        label_filenames = [label_file_name]
    else:  # split case
        label_filenames = []
        for job_name, job in jobs.items():
            labels_json["images"], labels_json["annotations"] = _get_coco_images_and_annotations(
                {job_name: job},
                assets,
                cat_kili_id_to_coco_id,
                project_input_type,
                annotation_modifier,
                is_single_job=True,
            )
            label_file_name = output_dir / job_name / "labels.json"

            label_file_name.parent.mkdir(parents=True, exist_ok=True)
            with label_file_name.open("w") as outfile:
                json.dump(labels_json, outfile)
            label_filenames.append(label_file_name)

    return labels_json, label_filenames


def _get_coco_categories_with_mapping(
    jobs: Dict[str, Job], merged: bool
) -> Tuple[Dict[str, Dict[str, int]], List[CocoCategory]]:
    """_get_coco_categories_with_mapping.

    Get the mapping between a category name in Kili of a given job and the COCO category id, and
    also return the list of COCO categories.
    """
    if merged:
        mapping_cat_name_cat_kili_id: Dict[str, str] = {}
        cat_kili_id_to_coco_id: Dict[str, Dict[str, int]] = {}
        id_offset = 0
        for job_name, job in sorted(jobs.items(), key=lambda x: x[0]):
            content = dict(job["content"])
            job_cats: Dict[str, Any] = content["categories"]
            mapping_cat_name_cat_kili_id = {
                "/".join([job_name, cat["name"]]): cat_id for cat_id, cat in job_cats.items()
            }

            cat_kili_ids = sorted(mapping_cat_name_cat_kili_id.values())
            cat_kili_id_to_coco_id[job_name] = {
                str(category_id): i + id_offset for i, category_id in enumerate(cat_kili_ids)
            }
            id_offset += len(cat_kili_ids)

    else:
        assert (
            len(list(jobs.values())) == 1
        ), "When this method is called with merged = False, the jobs should only contain 1 job"
        job_name = next(iter(jobs.keys()))
        cats = next(iter(jobs.values()))["content"]["categories"]
        mapping_cat_name_cat_kili_id = {cat["name"]: cat_id for cat_id, cat in cats.items()}
        cat_kili_ids = list(mapping_cat_name_cat_kili_id.values())
        cat_kili_id_to_coco_id = {
            job_name: {str(category_id): i for i, category_id in enumerate(cat_kili_ids)}
        }

    return cat_kili_id_to_coco_id, _get_coco_categories(cat_kili_id_to_coco_id, merged)


# pylint: disable=too-many-locals
def _get_coco_images_and_annotations(
    jobs: Dict[str, Job],
    assets: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, Dict[str, int]],
    project_input_type: str,
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    if project_input_type == "IMAGE":
        return _get_images_and_annotation_for_images(
            jobs, assets, cat_kili_id_to_coco_id, annotation_modifier, is_single_job
        )

    if project_input_type == "VIDEO":
        return _get_images_and_annotation_for_videos(
            jobs, assets, cat_kili_id_to_coco_id, annotation_modifier, is_single_job
        )

    raise NotImplementedError(f"No conversion to COCO possible for input type {project_input_type}")

def _get_images_and_annotation_for_images(
    jobs: Dict[str, Job],
    assets: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, Dict[str, int]],
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    coco_images = []
    coco_annotations = []
    annotation_offset = 0

    for asset_i, asset in enumerate(assets):
        width, height = get_image_dimensions(asset)
        if Path(asset["content"]).is_file():
            filename = str(DATA_SUBDIR + "/" + Path(asset["content"]).name)
        else:
            filename = str(DATA_SUBDIR + "/" + asset["externalId"])
        coco_image = CocoImage(
            id=asset_i,
            license=0,
            file_name=filename,
            height=height,
            width=width,
            date_captured=None,
        )
        rotation_val = 0
        if "ROTATION_JOB" in asset["latestLabel"]["jsonResponse"]:
            rotation_val = asset["latestLabel"]["jsonResponse"]["ROTATION_JOB"]["rotation"]
        coco_images.append(coco_image)
        if is_single_job:
            assert len(list(jobs.keys())) == 1
            job_name = next(iter(jobs.keys()))

            if job_name not in asset["latestLabel"]["jsonResponse"]:
                coco_img_annotations = []
                # annotation offset is unchanged
            else:
                coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                    asset["latestLabel"]["jsonResponse"][job_name]["annotations"],
                    cat_kili_id_to_coco_id[job_name],
                    annotation_offset,
                    coco_image,
                    annotation_modifier=annotation_modifier,
                    rotation=rotation_val,
                )
            coco_annotations.extend(coco_img_annotations)
        else:
            for job_name in jobs:
                if job_name not in asset["latestLabel"]["jsonResponse"]:
                    continue
                    # annotation offset is unchanged

                coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                    asset["latestLabel"]["jsonResponse"][job_name]["annotations"],
                    cat_kili_id_to_coco_id[job_name],
                    annotation_offset,
                    coco_image,
                    annotation_modifier=annotation_modifier,
                    rotation=rotation_val,
                )
                coco_annotations.extend(coco_img_annotations)
    return coco_images, coco_annotations


def _get_images_and_annotation_for_videos(
    jobs: Dict[str, Job],
    assets: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, Dict[str, int]],
    annotation_modifier: Optional[CocoAnnotationModifier],
    is_single_job: bool,
) -> Tuple[List[CocoImage], List[CocoAnnotation]]:
    coco_images = []
    coco_annotations = []
    annotation_offset = 0

    for asset in assets:
        nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
        leading_zeros = len(str(nbr_frames))

        width = height = 0
        frame_ext = ""
        # jsonContent with frames
        if isinstance(asset["jsonContent"], list) and Path(asset["jsonContent"][0]).is_file():
            width, height = get_frame_dimensions(asset)
            frame_ext = Path(asset["jsonContent"][0]).suffix

        # video with shouldUseNativeVideo set to True (no frames available)
        elif Path(asset["content"]).is_file():
            width, height = get_video_dimensions(asset)
            cut_video(asset["content"], asset, leading_zeros, Path(asset["content"]).parent)
            frame_ext = ".jpg"

        else:
            raise FileNotFoundError(f"Could not find frames or video for asset {asset}")

        for frame_i, (frame_id, json_response) in enumerate(
            asset["latestLabel"]["jsonResponse"].items()
        ):
            frame_name = f'{asset["externalId"]}_{str(int(frame_id)+1).zfill(leading_zeros)}'
            coco_image = CocoImage(
                id=frame_i + len(assets),  # add offset to avoid duplicate ids
                license=0,
                file_name=str(DATA_SUBDIR + "/" + f"{frame_name}{frame_ext}"),
                height=height,
                width=width,
                date_captured=None,
            )
            coco_images.append(coco_image)
            if is_single_job:
                job_name = next(iter(jobs.keys()))
                if job_name not in json_response:
                    coco_img_annotations = []
                    # annotation offset is unchanged
                else:
                    coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                        json_response[job_name]["annotations"],
                        cat_kili_id_to_coco_id[job_name],
                        annotation_offset,
                        coco_image,
                        annotation_modifier,
                    )
                coco_annotations.extend(coco_img_annotations)
            else:
                for job_name in jobs:
                    if job_name not in asset["latestLabel"]["jsonResponse"]:
                        continue
                    coco_img_annotations, annotation_offset = _get_coco_image_annotations(
                        json_response[job_name]["annotations"],
                        cat_kili_id_to_coco_id[job_name],
                        annotation_offset,
                        coco_image,
                        annotation_modifier,
                    )

                    coco_annotations.extend(coco_img_annotations)

    return coco_images, coco_annotations


def _get_coco_image_annotations(
    annotations_: List[Dict],
    cat_kili_id_to_coco_id: Dict[str, int],
    annotation_offset: int,
    coco_image: CocoImage,
    annotation_modifier: Optional[CocoAnnotationModifier],
    rotation: int = 0,
) -> Tuple[List[CocoAnnotation], int]:
    coco_annotations = []

    annotation_j = annotation_offset

    for annotation in annotations_:  # we do not use enumerate as some annotations may be empty
        annotation_j += 1

        if not annotation:
            print("continue")
            continue
        bounding_poly = annotation["boundingPoly"]
        area, bbox, polygons = _get_coco_geometry_from_kili_bpoly(
            bounding_poly, coco_image["width"], coco_image["height"], rotation
        )
        if len(polygons[0]) < 6:  # twice the number of vertices
            print("A polygon must contain more than 2 points. Skipping this polygon...")
            continue
        if bbox[2] == 0 and bbox[3] == 0:
            print("An annotation with zero dimensions has been ignored.")
            continue

        categories = annotation["categories"]
        coco_annotation = CocoAnnotation(
            id=annotation_j,
            image_id=coco_image["id"],
            category_id=cat_kili_id_to_coco_id[categories[0]["name"]],
            bbox=bbox,
            # Objects have only one connected part.
            # But a type of object can appear several times on the same image.
            # The limitation of the single connected part comes from Kili.
            segmentation=polygons,
            area=area,
            iscrowd=0,
        )

        if annotation_modifier:
            coco_annotation = annotation_modifier(
                dict(coco_annotation), dict(coco_image), annotation
            )

        coco_annotations.append(coco_annotation)
    return coco_annotations, annotation_j




def _get_coco_geometry_from_kili_bpoly(
    bounding_poly: List[Dict], asset_width: int, asset_height: int, rotation_angle: int
):
    normalized_vertices = bounding_poly[0]["normalizedVertices"]
    vertices_before_rotate = reverse_rotation_vertices(normalized_vertices, rotation_angle)
    p_x = [float(vertice["x"]) * asset_width for vertice in vertices_before_rotate]
    p_y = [float(vertice["y"]) * asset_height for vertice in vertices_before_rotate]
    poly_vertices = [(float(x), float(y)) for x, y in zip(p_x, p_y)]
    x_min, y_min = round(min(p_x)), round(min(p_y))
    x_max, y_max = round(max(p_x)), round(max(p_y))
    bbox_width, bbox_height = x_max - x_min, y_max - y_min
    area = round(_get_shoelace_area(p_x, p_y))
    polygons = [[p for vertice in poly_vertices for p in vertice]]

    # Compute and remove negative area
    if len(bounding_poly) > 1:
        for negative_bounding_poly in bounding_poly[1:]:
            negative_normalized_vertices = negative_bounding_poly["normalizedVertices"]
            negative_vertices_before_rotate = reverse_rotation_vertices(
                negative_normalized_vertices, rotation_angle
            )
            np_x = [
                float(negative_vertice["x"]) * asset_width
                for negative_vertice in negative_vertices_before_rotate
            ]
            np_y = [
                float(negative_vertice["y"]) * asset_height
                for negative_vertice in negative_vertices_before_rotate
            ]
            area -= _get_shoelace_area(np_x, np_y)
            poly_negative_vertices = [(float(x), float(y)) for x, y in zip(np_x, np_y)]
            polygons.append([p for vertice in poly_negative_vertices for p in vertice])
    bbox = [int(x_min), int(y_min), int(bbox_width), int(bbox_height)]
    return area, bbox, polygons


def _get_coco_categories(cat_kili_id_to_coco_id, merged) -> List[CocoCategory]:
    categories_coco: List[CocoCategory] = []
    for job_name, mapping in sorted(cat_kili_id_to_coco_id.items(), key=lambda x: x[0]):
        for cat_kili_id, cat_coco_id in sorted(mapping.items(), key=lambda x: x[1]):
            categories_coco.append(
                {
                    "id": cat_coco_id,
                    "name": cat_kili_id if not merged else f"{job_name}/{cat_kili_id}",
                    "supercategory": job_name,
                }
            )

    return categories_coco


# Shoelace formula, implementation from https://stackoverflow.com/a/30408825.
def _get_shoelace_area(x: List[float], y: List[float]):
    # Split self intersecting polygon into multiple polygons to compute area safely.
    polygon = Polygon(np.c_[x, y])
    polygons = polygonize(make_valid(polygon))

    area = 0

    for poly in polygons:
        p_xx, p_yy = poly.exterior.coords.xy
        p_x = p_xx.tolist()
        p_y = p_yy.tolist()
        area += 0.5 * np.abs(np.dot(p_x, np.roll(p_y, 1)) - np.dot(p_y, np.roll(p_x, 1)))

    return area



from enum import Enum
from typing import Any, Dict, List, Literal, NamedTuple, Optional, TypedDict

InputType = Literal["IMAGE", "LLM_INSTR_FOLLOWING", "LLM_RLHF", "PDF", "TEXT", "VIDEO"]
MLTask = Literal["CLASSIFICATION", "NAMED_ENTITIES_RECOGNITION", "OBJECT_DETECTION"]

class JobCategory(NamedTuple):
    """Contains information for a category."""

    category_name: str
    id: int
    job_id: str

class JobTool(str, Enum):
    """List of tools."""

    MARKER = "marker"
    POLYGON = "polygon"
    POLYLINE = "polyline"
    POSE = "pose"
    RANGE = "range"
    RECTANGLE = "rectangle"
    SEMANTIC = "semantic"
    VECTOR = "vector"

class Job(TypedDict):
    """Contains job settings."""

    content: Any
    instruction: str
    isChild: bool
    tools: List[JobTool]
    mlTask: MLTask
    models: Any  # example: {"interactive-segmentation": {"job": "SEMANTIC_JOB_MARKER"}},
    isVisible: bool
    required: int
    isNew: bool

class ProjectDict(TypedDict):
    description: str
    id: str
    inputType: InputType
    jsonInterface: Optional[Dict]
    organizationId: str
    title: str


class ChatItemRole(str, Enum):
    """Enumeration of the supported chat item role."""

    ASSISTANT = "ASSISTANT"
    USER = "USER"
    SYSTEM = "SYSTEM"
    
class ChatItem(TypedDict):
    """Dict that represents a ChatItem."""

    id: str
    content: str
    createdAt: Optional[str]
    externalId: str
    modelId: Optional[str]
    modelName: Optional[str]
    role: ChatItemRole


class ConversationLabel(TypedDict):
    """Dict that represents a ConversationLabel."""

    completion: Optional[Dict]
    conversation: Optional[Dict]
    round: Optional[Dict]


class Conversation(TypedDict):
    """Dict that represents a Conversation."""

    chatItems: List[ChatItem]
    externalId: Optional[str]
    label: Optional[ConversationLabel]
    labeler: Optional[str]
    metadata: Optional[dict]

class ExportLLMItem(TypedDict):
    """LLM asset chat part."""

    role: str
    content: str
    id: Optional[str]
    chat_id: Optional[str]
    model: Optional[str]

class JobLevel:
    ROUND = "round"
    CONVERSATION = "conversation"
    COMPLETION = "completion"
